'use strict';

var R = require('ramda');

var _require = require('creed'),
    resolve = _require.resolve,
    all = _require.all;

var _require2 = require('silverscreen'),
    infoForVideoWithURL = _require2.infoForVideoWithURL;

var parseElement = require('./parseElement');
var renderMarkdown = require('./renderMarkdown');
var htmlElement = require('./htmlElement');
var createRoutesForHTML = require('./createRoutesForHTML');
var promiseEmbedInfoForURL = require('./promiseEmbedInfoForURL');

var itemNameRegex = R.pipe(R.test, R.propSatisfies(R.__, 'name'));
var isPageList = itemNameRegex(/\B#path:/);
var isHomePageList = itemNameRegex(/\B#path:\s*\/\B/);

var itemHasName = R.propEq('name');
var isStrictMetaCard = itemHasName('#meta');
var isCSSCard = itemNameRegex(/\B#css:/);
var isLanguageCard = itemNameRegex(/\B#language:/);
var isAboveContentCard = itemHasName('#above');
var isBelowContentCard = itemHasName('#below');

var groupCards = R.pipe(R.groupBy(R.cond([[R.anyPass([isStrictMetaCard, isCSSCard]), R.always('metaCards')], [isAboveContentCard, R.always('aboveCards')], [isBelowContentCard, R.always('belowCards')], [isLanguageCard, R.always('languageCards')], [R.T, R.always('contentCards')]])), R.merge({
    metaCards: [],
    aboveCards: [],
    belowCards: [],
    languageCards: [],
    contentCards: []
}));

function cardsForID(listID, allCards) {
    return allCards.filter(function (card) {
        return card.idList === listID && !card.closed;
    });
}

var filterDownloadAttachments = R.filter(R.allPass([R.propEq('isUpload', true), R.propSatisfies(R.isEmpty, 'previews')]));

var filterLinkAttachmentsURLs = R.pipe(R.filter(R.propEq('isUpload', false)), R.pluck('url'));

function resolveContent(content, defaultValue) {
    var transformer = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : R.identity;

    if (!content) {
        return defaultValue;
    }

    if (content == true) {
        return true;
    }

    if (content.text) {
        return transformer(content.text);
    }

    return defaultValue;
}

var standardMode = {
    sectionTag: 'article',
    itemTag: null
};

var articlesMode = {
    sectionTag: 'section',
    itemTag: 'article'
};

var navMode = {
    sectionTag: 'nav',
    itemTag: null
};

function renderCards(cards) {
    var _ref = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
        _ref$mode = _ref.mode,
        mode = _ref$mode === undefined ? {} : _ref$mode,
        _ref$path = _ref.path,
        path = _ref$path === undefined ? '/' : _ref$path,
        title = _ref.title;

    var sectionAttributes = {};
    var itemAttributes = {};

    var _cards$reduce = cards.reduce(function (combined, _ref2) {
        var name = _ref2.name,
            attachments = _ref2.attachments,
            desc = _ref2.desc,
            linkURL = _ref2.linkURL,
            videoInfo = _ref2.videoInfo,
            embedInfo = _ref2.embedInfo;

        // TODO: use element: { text, tags }
        var _parseElement = parseElement(name),
            text = _parseElement.text,
            tags = _parseElement.tags;

        var draft = resolveContent(tags.draft, false);
        // Skip drafts
        if (draft) {
            return combined;
        }

        // #nav
        if (tags.nav) {
            mode = navMode;
            itemAttributes = {};

            var className = resolveContent(tags.class) || '';
            if (tags.primary) {
                className += ' primary';
            }

            className = className.trim();
            if (className.length > 0) {
                sectionAttributes.class = className;
            }
            return combined;
        }
        // #article
        else if (tags.article) {
                mode = articlesMode;
                itemAttributes = {};
                var _className = resolveContent(tags.class);
                if (_className) {
                    itemAttributes.class = _className;
                }
                return combined;
            }

        var imagesBefore = false;
        var itemTag = mode.itemTag;

        // Slug and child path
        var slug = resolveContent(tags.slug);
        var childPath = void 0;
        if (slug) {
            // Combine path with slug
            childPath = R.pipe(R.concat(R.__, '/'), R.concat(R.__, slug), R.replace('//', '/'))(path);

            if (!itemTag) {
                itemTag = 'article';
            }
        }

        // Image scale factor
        var scaleFactor = resolveContent(tags.scalefactor, 1.0, parseFloat);
        // Attached images
        var imagesHTML = R.chain(function (_ref3) {
            var previews = _ref3.previews;

            if (!previews) {
                return [];
            }

            var preview = R.last(previews);
            if (!preview) {
                return [];
            }

            return htmlElement('img', { src: preview.url, width: preview.width / scaleFactor, height: preview.height / scaleFactor });
        }, attachments).join('\n');

        var output = void 0;

        if (tags.primary) {
            output = '';
            title = text;
        } else if (text.length > 0) {
            output = text;

            var tagName = 'h2'; // #secondary by default
            var outerTagName = void 0;
            var classes = [];

            var baseClass = resolveContent(tags.class);
            if (baseClass) {
                classes.push(baseClass);
            }

            if (tags.figure) {
                itemTag = 'figure';
                outerTagName = 'figcaption';
                imagesBefore = true;
            } else if (tags.secondary) {
                tagName = 'h2';
            } else if (tags.tertiary) {
                tagName = 'h3';
            }
            // TODO: remove
            else if (tags.text) {
                    tagName = 'p';
                }

            if (childPath) {
                output = htmlElement('a', {
                    href: childPath
                }, output);
            }

            output = htmlElement(tagName, {
                class: classes.join(' ')
            }, output);
        } else {
            output = '';
        }

        if (videoInfo) {
            // Add embedded video
            output += videoInfo.desktopSize.embedCode;
        } else if (embedInfo) {
            output += htmlElement('iframe', {
                src: embedInfo.url,
                width: embedInfo.width || 300,
                height: embedInfo.height || 200,
                frameborder: 0
            });
        }

        var descriptionHTML = '';
        if (desc && desc.length > 0) {
            if (tags.html) {
                // Raw unsafe HTML
                descriptionHTML = desc;
            } else {
                // Render Markdown
                descriptionHTML = renderMarkdown(desc);
            }
        }

        // Add images
        if (imagesBefore) {
            output = imagesHTML + output + descriptionHTML;
        } else {
            output = output + imagesHTML + descriptionHTML;
        }

        // Attachment links
        var downloadAttachments = filterDownloadAttachments(attachments);
        if (downloadAttachments.length > 0) {
            output += renderDownloadAttachmentList(downloadAttachments);
        }

        if (output.length > 0) {
            if (itemTag) {
                output = htmlElement(itemTag, itemAttributes, output) + '\n';
            }

            //const linkURL = resolveContent(tags.link)
            if (linkURL) {
                var _classes = [];
                if (tags.cta) {
                    _classes.push('cta');
                }
                // Change itemTag to 'a' instead?
                output = htmlElement('a', { href: linkURL, class: _classes.join(' ') }, output);
            }

            if (slug) {
                combined.children.push({
                    slug: slug,
                    path: childPath,
                    title: text,
                    html: output
                });
            }
        }

        combined.html += output + '\n';

        return combined;
    }, { html: '', children: [] }),
        html = _cards$reduce.html,
        children = _cards$reduce.children;

    if (mode.sectionTag) {
        html = htmlElement(mode.sectionTag, sectionAttributes, html) + '\n';
    }

    return { html: html, children: children, title: title };
}

function renderContentHTML(html, title) {
    var _ref4 = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {},
        _ref4$mainTag = _ref4.mainTag,
        mainTag = _ref4$mainTag === undefined ? 'main' : _ref4$mainTag,
        _ref4$mainClass = _ref4.mainClass,
        mainClass = _ref4$mainClass === undefined ? 'content' : _ref4$mainClass;

    if (title != null && title.length > 0) {
        var header = htmlElement('header', {}, htmlElement('h1', {}, title));
        html = '\n' + header + '\n' + html + '\n';
    }

    return htmlElement(mainTag, { class: mainClass }, html);
}

function renderContentCards(cards, _ref5) {
    var defaultTitle = _ref5.defaultTitle,
        path = _ref5.path;

    var _renderCards = renderCards(cards, { mode: standardMode, title: defaultTitle, path: path }),
        html = _renderCards.html,
        children = _renderCards.children,
        title = _renderCards.title;

    html = renderContentHTML(html, title);

    return {
        html: html,
        children: children,
        title: title
    };
}

function renderMetaCards(cards) {
    return cards.map(function (card) {
        if (isStrictMetaCard(card)) {
            // Raw HTML
            return card.desc;
        } else {
            // Convenience via tags
            var _ref6 = card.element || parseElement(card.name),
                text = _ref6.text,
                tags = _ref6.tags;

            if (tags.css) {
                var cssURL = resolveContent(tags.css);
                if (R.is(String, cssURL) && cssURL.length > 0) {
                    return htmlElement('link', { rel: 'stylesheet', href: cssURL });
                } else if (card.desc.length > 0) {
                    return htmlElement('style', {}, card.desc);
                } else {
                    return '';
                }
            } else {
                return '<!-- Unknown \'' + card.name + '\' -->';
            }
        }
    }).join('\n');
}

var renderDownloadAttachmentList = R.pipe(R.map(R.pipe(renderDownloadAttachmentLink, function (html) {
    return htmlElement('li', {}, html);
})), R.join("\n"), function (html) {
    return htmlElement('ul', {}, html);
});

function renderDownloadAttachmentLink(_ref7) {
    var name = _ref7.name,
        url = _ref7.url;

    return htmlElement('a', { href: url }, name);
}

function renderSiteNavigation(_ref8) {
    var pageLists = _ref8.pageLists;

    return htmlElement('nav', {}, pageLists.map(function (list) {
        var name = list.name;

        var _parseElement2 = parseElement(name),
            text = _parseElement2.text,
            tags = _parseElement2.tags;

        return htmlElement('h2', {}, htmlElement('a', {
            href: resolveContent(tags.path)
        }, text));
    }).join("\n"));
}

function routesForRedirectCards(cards) {
    return cards.reduce(function (routes, _ref9) {
        var name = _ref9.name;

        var _parseElement3 = parseElement(name),
            text = _parseElement3.text,
            tags = _parseElement3.tags;

        var fromPath = resolveContent(tags.from);
        var toPath = resolveContent(tags.to);
        if (fromPath && toPath) {
            routes.push({
                method: 'GET',
                path: fromPath,
                handler: function handler(request, reply) {
                    reply.redirect(toPath);
                }
            });
        }

        return routes;
    }, []);
}

var promiseEnhancedCards = R.pipe(R.map(function (card) {
    var _parseElement4 = parseElement(card.name),
        text = _parseElement4.text,
        tags = _parseElement4.tags;

    var linkURL =
    // #link
    resolveContent(tags.link) ||
    // First attached link’s URL
    filterLinkAttachmentsURLs(card.attachments)[0];

    var promises = [resolve(card), resolve({
        element: { text: text, tags: tags },
        linkURL: linkURL
    })];

    var videoURL = resolveContent(tags.video);
    if (videoURL) {
        promises.push(infoForVideoWithURL(videoURL, {
            embedding: {
                width: 500,
                aspectRatio: 9 / 16
            }
        }).then(R.objOf('videoInfo')));
    }

    var embed = resolveContent(tags.embed, false);
    if (embed && !!linkURL) {
        promises.push(promiseEmbedInfoForURL(linkURL).then(R.objOf('embedInfo')));
    }

    return all(promises).map(R.mergeAll);
}), all);

function routesForTrelloData(_ref10) {
    var siteTitle = _ref10.name,
        lists = _ref10.lists,
        allCards = _ref10.cards;

    // Ignore archived cards
    lists = lists.filter(function (list) {
        return !list.closed;
    });
    // Lists that have #path
    var pageLists = lists.filter(isPageList);
    // Home
    var hasHomePageList = R.any(isHomePageList, lists);

    // Global
    var globalLists = lists.filter(function (_ref11) {
        var name = _ref11.name;
        return name === '#all';
    });
    var globalCardsGrouped = R.chain(function (_ref12) {
        var listID = _ref12.id;
        return groupCards(cardsForID(listID, allCards));
    }, globalLists);
    // Meta
    var globalMetaCards = R.chain(function (_ref13) {
        var metaCards = _ref13.metaCards;
        return metaCards;
    }, globalCardsGrouped);
    // Above
    var globalAboveCards = R.chain(function (_ref14) {
        var aboveCards = _ref14.aboveCards;
        return aboveCards;
    }, globalCardsGrouped);

    var _renderCards2 = renderCards(globalAboveCards),
        globalAboveHTML = _renderCards2.html;
    // Below


    var globalBelowCards = R.chain(function (_ref15) {
        var belowCards = _ref15.belowCards;
        return belowCards;
    }, globalCardsGrouped);

    var _renderCards3 = renderCards(globalBelowCards),
        globalBelowHTML = _renderCards3.html;
    // Nav


    var showBelowNav = !hasHomePageList;
    var globalNavHTML = showBelowNav ? renderSiteNavigation({ pageLists: pageLists }) : '';

    var getCombinedMetaCards = function getCombinedMetaCards(additionalCards) {
        var combinedMetaCards = globalMetaCards.concat(additionalCards);
        if (!R.any(isCSSCard, combinedMetaCards)) {
            combinedMetaCards.push({
                name: '#meta',
                desc: require('./default/styleElement')()
            });
        }
        return combinedMetaCards;
    };

    var routes = lists.reduce(function (routes, _ref16) {
        var listName = _ref16.name,
            listID = _ref16.id;

        var cards = cardsForID(listID, allCards);

        // Redirects
        if (listName === '#redirects') {
            routes.push.apply(routes, routesForRedirectCards(cards));
            return routes;
        }

        // Page with path

        var _parseElement5 = parseElement(listName),
            title = _parseElement5.text,
            tags = _parseElement5.tags;

        var path = resolveContent(tags.path);
        if (path == null || path.length === 0) {
            return routes;
        }

        var amp = resolveContent(tags.amp, false, Boolean);

        var bodyClasses = (resolveContent(tags.class) || '').split(' ');

        var _groupCards = groupCards(cards),
            metaCards = _groupCards.metaCards,
            contentCards = _groupCards.contentCards;

        var _renderContentCards = renderContentCards(contentCards, { defaultTitle: title, path: path }),
            contentHTML = _renderContentCards.html,
            children = _renderContentCards.children;

        var bodyHTML = globalAboveHTML + contentHTML + globalNavHTML + globalBelowHTML;
        var metaHTML = renderMetaCards(getCombinedMetaCards(metaCards));

        routes.push.apply(routes, createRoutesForHTML({
            path: path,
            htmlOptions: {
                title: title,
                metaHTML: metaHTML,
                bodyClasses: bodyClasses,
                bodyHTML: bodyHTML
            },
            amp: amp
        }));

        if (children.length > 0) {
            children.forEach(function (_ref17) {
                var slug = _ref17.slug,
                    path = _ref17.path,
                    title = _ref17.title,
                    contentHTML = _ref17.html;

                var bodyHTML = renderContentHTML(globalAboveHTML + contentHTML + globalNavHTML + globalBelowHTML);

                routes.push.apply(routes, createRoutesForHTML({
                    path: path,
                    htmlOptions: {
                        title: title,
                        metaHTML: metaHTML,
                        bodyClasses: bodyClasses,
                        bodyHTML: bodyHTML
                    },
                    amp: amp
                }));
            });
        }

        return routes;
    }, []);

    if (!hasHomePageList) {
        routes.push.apply(routes, createRoutesForHTML({
            path: '/',
            htmlOptions: {
                title: 'Home',
                metaHTML: renderMetaCards(getCombinedMetaCards([])),
                bodyHTML: renderContentHTML(globalNavHTML, siteTitle)
            }
        }));
    }

    return routes;
}

module.exports = {
    routesForTrelloData: routesForTrelloData,
    promiseEnhancedTrelloCards: promiseEnhancedCards
};